/****************************************************************
*
* AnomalousStateFinder.cc	XMM Science Analysis System (SAS)
*
*
* Class containing routines to identify and tag anomalous states
* according to Kuntz and Snowden (2008) A&A 478, 575--596.
*
* (c) 2008 ESA - Written by Hubert Chen, CEA Saclay
*
* $Id: AnomalousStateFinder.cc,v 1.0 2009/04/16 09:38:32 cmhchen Exp $
*
****************************************************************/

#include "AnomalousStateFinder.h"

//------------------------------------------------------------------------------
/** Constructor that takes a filename string as input. */
AnomalousStateFinder::AnomalousStateFinder(const string filename)
  : _filename(filename) {
// filename |-> DataSet.
  // Open file.
  _dataSetPtr = dataSetServer->open( filename, Dal::Modify );
  init();
} // AnomalousStateFinder::AnomalousStateFinder(const string filename)

//------------------------------------------------------------------------------
/** Constructor that takes a pointer to a DataSet instance as input. */
AnomalousStateFinder::AnomalousStateFinder(DataSet * dataSetPtr)
  : _filename("") {
  _dataSetPtr = dataSetPtr;
  init();
} // AnomalousStateFinder::AnomalousStateFinder(DataSet * dataSetPtr)

//------------------------------------------------------------------------------
/** Handle tasks common to all constructors. */
void AnomalousStateFinder::init() {
  // Set the state of the calServer.
  // Must do this for calServer.instrument()->* to function properly.
  calServer.setState(*_dataSetPtr);
  const XmmInstrument * instrumentPtr = calServer.instrument();

  Instrument id = instrumentPtr->id();
  if ( (id != Xmm::EMOS1) && (id != Xmm::EMOS2) )
    errstr	<< "Instrument type " << instrumentPtr << " not recognised."
		<< Sas::error("unrecognisedInstrument");

  _CCDdataPtrArray = new (CCDdata *)[instrumentPtr->nCcds()];
  for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++) {
    _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]
	= new CCDdata(_dataSetPtr, ccdnr);
  } // for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++)

  _eventsRead = false;

  // Set up selectlib evaluator for computing isCornerPixel().
  _cornerEvaluatorPtr = new Evaluator( cornerExpression() );
  _cornerEvalSyms = _cornerEvaluatorPtr->numericIdentifiers();
  // Check _cornerEvalSyms for variables DETX and DETY.
  _detXptrPos = -1;
  _detYptrPos = -1;
  for (unsigned i=0u; i < _cornerEvalSyms.name.size(); i++)
    if (_cornerEvalSyms.name[i] == "DETX") _detXptrPos = i;
    else
    if (_cornerEvalSyms.name[i] == "DETY") _detYptrPos = i;
    else
      Sas::error("invalidSelectionExpression",
	"`Corner pixel' expression cannot contain"
	" variables other than DETX and DETY.");
} // void AnomalousStateFinder::init()

//------------------------------------------------------------------------------
AnomalousStateFinder::~AnomalousStateFinder() {
  delete _cornerEvaluatorPtr;

  for (unsigned index=EMOScam::CCD_LAST - EMOScam::CCD_FIRST; index > 0u; index--) {
    delete _CCDdataPtrArray[index]; _CCDdataPtrArray[index] = 0;
  } // for (unsigned index=EMOScam::CCD_LAST - EMOScam::CCD_FIRST; index > 0u; index--)
  delete [] _CCDdataPtrArray; _CCDdataPtrArray = 0;

  // Close file.
  if (_dataSetPtr != NULL)
    dataSetServer->close(_dataSetPtr);

//errstr<<"AnomalousStateFinder deleted"<<Sas::message(Msg::App,Msg::Always);
} // AnomalousStateFinder::~AnomalousStateFinder()

//------------------------------------------------------------------------------
/** Main public interface with a filename string as input.
    Just open the file for modification and call the version
    with a DataSet* as input to do the real work. */
void AnomalousStateFinder::tagAnomalous(const string filename) {
  DataSet * dataSetPtr = dataSetServer->open( filename, Dal::Modify );
  tagAnomalous(dataSetPtr);
  dataSetServer->close(dataSetPtr);
} // void AnomalousStateFinder::tagAnomalous(const string filename)

//------------------------------------------------------------------------------
/** Main public interface with a DataSet* as input.
    Instantiate AnomalousStateFinder.
    Call writeAttributes() to do the actual work. */
void AnomalousStateFinder::tagAnomalous(DataSet * dataSetPtr) {
  AnomalousStateFinder * anomalousStateFinderPtr
	= new AnomalousStateFinder(dataSetPtr);

  anomalousStateFinderPtr->writeAttributes();
/*
  for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++) {
    if ( anomalousStateFinderPtr->isAnomalous(ccdnr) )
	// For now, print to STDERR.
	errstr<<"CCD "<<ccdnr<<" IS anomalous."<<Sas::message(Msg::App,Msg::Always);
    else
	errstr<<"CCD "<<ccdnr<<" is NOT anomalous."<<Sas::message(Msg::App,Msg::Always);
  } // for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++)
*/

  delete anomalousStateFinderPtr; anomalousStateFinderPtr = 0;
} // void AnomalousStateFinder::tagAnomalous(DataSet * dataSetPtr)

//------------------------------------------------------------------------------
/** Wrapper to call writeAttribute(Ccd) for each CCD except the central CCD. */
void AnomalousStateFinder::writeAttributes() {
  for (Ccd ccdnr=EMOScam::CCD_FIRST; ccdnr <= EMOScam::CCD_LAST; ccdnr++)
    if (ccdnr != EMOScam::CCD_CENTRAL)
      writeAttribute(ccdnr);
} // void AnomalousStateFinder::writeAttributes()

//------------------------------------------------------------------------------
/** Write anomalous-state attribute for ccdnr to _dataSet
    according to the return value of isAnomalous(ccdnr). */
void AnomalousStateFinder::writeAttribute(Ccd ccdnr) {
  // Construct the various input fields of an Attribute.
  const string	name    = attributeName   (ccdnr);
  const string	comment = attributeComment(ccdnr);
  int		value   = isAnomalous(ccdnr)? 1 : 0;

  // Check if attribute already exists in *_dataSetPtr.
  if ( _dataSetPtr->hasAttribute(name) ) {
	// Braces needed to avoid ambiguity of else clause.
    // Exit with error.
    if (_dataSetPtr->attribute(name)->asInt() != value)
      Sas::error(name + " mismatch.",
		  "Existing value of attribute " + name
		+ " does not match newly calculated value.");
    // Otherwise, do nothing.
  } // if _dataSetPtr->hasAttribute(name)
  else
    // Write new attribute to DataSet.
    *( _dataSetPtr->addAttribute(name, comment) ) = value;
} // void AnomalousStateFinder::writeAttribute(Ccd ccdnr)

//------------------------------------------------------------------------------
/** True if the specified CCD is in an anomalous state. */
bool AnomalousStateFinder::isAnomalous(const Ccd ccdnr) {
  if (!_eventsRead) readEvents();
  return _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]->isAnomalous();
} // bool AnomalousStateFinder::isAnomalous(const Ccd ccdnr)

//------------------------------------------------------------------------------
/** Subroutine to provide abstraction for the `corner pixel' criteria. */
const string AnomalousStateFinder::cornerExpression() {
  // Code to interface with cal from RDS.
  GeometryDataServer *ct = 0; ct = calServer.getAtom(ct);
  vector<string> corners = ct->corners();

  //ostringstream outStream(corners[0]); // does not work and I do not know why.
  ostringstream outStream;
  outStream << corners[0];
  for (unsigned i=1u; i < corners.size(); i++)
    outStream << " && " << corners[i];
  const string & outString = outStream.str();
  return outString;
} // AnomalousStateFinder::cornerExpression()

//------------------------------------------------------------------------------
/** Return the selectlib expression that decides if the specified CCD is
    in an anomalous state.  The returned value contains no variable
    string other than "H" (hardness ratio) and "R" (count rate)
    as per Table 1 of Kuntz and Snowden (2008). */
const string AnomalousStateFinder::anomalousExpression(Ccd ccdnr) {
  return _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]->anomalousExpression();
} // AnomalousStateFinder::anomalousExpression(Ccd ccdnr)

//------------------------------------------------------------------------------
/** Format string with which to construct the attribute names. */
const char * AnomalousStateFinder::attributeNameFormat = "LENOIS0%d";

//------------------------------------------------------------------------------
/** Format string with which to construct the attribute comments. */
const char * AnomalousStateFinder::attributeCommentFormat
				= "Does CCD %d suffer from low-energy noise?";

//------------------------------------------------------------------------------
/** Name of the attribute to write to DataSet. */
const string AnomalousStateFinder::attributeName(Ccd ccdnr) {
  char * sprintfOutput = new char[80];	// maximum length of a FITS header card
  assert(ccdnr >= EMOScam::CCD_FIRST);
  assert(ccdnr <= EMOScam::CCD_LAST );
  sprintf(sprintfOutput, attributeNameFormat, ccdnr);
  string attributeName(sprintfOutput);
  delete [] sprintfOutput;
  return attributeName;
} // const string AnomalousStateFinder::attributeName(Ccd ccdnr)

//------------------------------------------------------------------------------
/** Comment for the attribute to write to DataSet. */
const string AnomalousStateFinder::attributeComment(Ccd ccdnr) {
  char * sprintfOutput = new char[80];	// maximum length of a FITS header card
  assert(ccdnr >= EMOScam::CCD_FIRST);
  assert(ccdnr <= EMOScam::CCD_LAST );
  sprintf(sprintfOutput, attributeCommentFormat, ccdnr);
  string attributeComment(sprintfOutput);
  delete [] sprintfOutput;
  return attributeComment;
} // const string AnomalousStateFinder::attributeComment(Ccd ccdnr)

//------------------------------------------------------------------------------
/** Constructor.
    instrument must be either Xmm::EMOS1 or XMM::EMOS2.
    DataSet must be of instrument Xmm::EMOS1 or XMM::EMOS2.
    ccdnr is the CCD number of this CCDdata.
    DataSet is defined in dal/interface/Dal.h . */
AnomalousStateFinder::CCDdata::CCDdata(const DataSet * dataSetPtr, const Ccd ccdnr)
  : _dataSetPtr(dataSetPtr),
    _ccdnr(ccdnr) {
  _instrument = calServer.instrument()->id();
  // Set private variables negative to signify invalidity.
  _GTIsum = -1.0;
  for (unsigned i=0u; i < N_BANDS; i++) {
    _counters[i] = 0ul;
    _counters_last_accessed[i] = -1ul; // for comparing with _counters[].
  } // for (unsigned i=0u; i < N_BANDS; i++)

  // Set up selectlib evaluator for computing isAnomalous().
  _evaluatorPtr = new Evaluator( anomalousExpression() );
  _evalSyms = _evaluatorPtr->numericIdentifiers();
  // Check _evalSyms for variables H and R.
  _evalHptrPos = -1;
  _evalRptrPos = -1;
  for (unsigned i=0u; i < _evalSyms.name.size(); i++)
    if (_evalSyms.name[i] == "H") _evalHptrPos = i;
    else
    if (_evalSyms.name[i] == "R") _evalRptrPos = i;
    else
      Sas::error("invalidSelectionExpression",
	"Anomalous expression cannot contain variables other than H and R.");

  // Get boundaries of energy bands from cal once and for all.
  // Energies are storied in CCFs in keVs.
  // Code to interface with cal from RDS.
  calServer.state()->ccd()->set(_ccdnr);
  ScreenThresholdServer *st = 0; st = calServer.getAtom(st);
  for (unsigned i=0u; i < N_BANDS; i++)
    switch (i) {
      case RATEBAND:
	_minEnergy[i] = (int)( 1000 * st->ratebandmin() );
	_maxEnergy[i] = (int)( 1000 * st->ratebandmax() );
	break; // switch (i)
      case SOFTBAND:
	_minEnergy[i] = (int)( 1000 * st->softbandmin() );
	_maxEnergy[i] = (int)( 1000 * st->softbandmax() );
	break; // switch (i)
      case HARDBAND:
	_minEnergy[i] = (int)( 1000 * st->hardbandmin() );
	_maxEnergy[i] = (int)( 1000 * st->hardbandmax() );
	break; // switch (i)
      default:
	Sas::error("unrecognisedEnergyBand", "Unrecognised energy band.");
    } // switch (i)
} // AnomalousStateFinder::CCDdata::CCDdata(const DataSet * dataSetPtr, const Ccd ccdnr)

//------------------------------------------------------------------------------
AnomalousStateFinder::CCDdata::~CCDdata() {
  delete _evaluatorPtr; _evaluatorPtr = 0;
//errstr<<"CCDdata "<<_ccdnr<<" deleted."<<Sas::message(Msg::App,Msg::Always);
} // AnomalousStateFinder::CCDdata::~CCDdata()

//------------------------------------------------------------------------------
const string AnomalousStateFinder::CCDdata::anomalousExpression() {
  if ( (_ccdnr < EMOScam::CCD_FIRST) || (_ccdnr > EMOScam::CCD_LAST) )
    errstr	<< "CCD number " << _ccdnr << " out of range."
		<< Sas::error("CCDNRoutOfRange.");

  // Code to interface with cal from RDS.
  calServer.state()->ccd()->set(_ccdnr);
  ScreenThresholdServer *st = 0; st = calServer.getAtom(st);
  const string & thresh = st->threshold();

  // Some complication: Need to remove any leading or trailing quotes.
  size_t first_real = thresh.find_first_not_of(" '\"\t");
  if (first_real == string::npos) return "false";
  size_t  last_real = thresh.find_last_not_of( " '\"\t");

  return thresh.substr( first_real, (last_real-first_real)+1 );
} // const string AnomalousStateFinder::CCDdata::anomalousExpression()

//------------------------------------------------------------------------------
bool AnomalousStateFinder::CCDdata::isInBand(int eVtestEnergy, EnergyBand band) const {
  return (_minEnergy[band] <= eVtestEnergy) && (eVtestEnergy < _maxEnergy[band]);
} // bool AnomalousStateFinder::CCDdata::isInBand(int eVtestEnergy, EnergyBand band) const

//------------------------------------------------------------------------------
/** Increment the event counter for the specified band. */
void AnomalousStateFinder::CCDdata::increment(const EnergyBand band, const unsigned long increment) {
  _counters[band] += increment;
} // void AnomalousStateFinder::CCDdata::increment(const EnergyBand band, const unsigned long increment = 1ul)

//------------------------------------------------------------------------------
/** Return the event count (to date) for the specified band. */
unsigned long AnomalousStateFinder::CCDdata::getCount(const EnergyBand band) const {
  return _counters[band];
} // unsigned long AnomalousStateFinder::CCDdata::getCount(const EnergyBand band) const

//------------------------------------------------------------------------------
/** Return the sum of the GTIs for this (Instrument,Ccd). */
double AnomalousStateFinder::CCDdata::getGTIsum() {
  if (_GTIsum < 0.0) _GTIsum = sumGTI();
  return _GTIsum;
} // double AnomalousStateFinder::CCDdata::getGTIsum()

//------------------------------------------------------------------------------
/** Return the hardness ratio, H = counts(HARDBAND) / counts(SOFTBAND).
    This method recomputes H if increment(...)
    was called after the most recent call of getH().
    If the soft-band count is null,
    a warning is issued, and -1 * hard-band count is returned. */
double AnomalousStateFinder::CCDdata::getH() {
  unsigned long softband_count = _counters[SOFTBAND];
  if (softband_count == 0ul) {
    errstr << "Soft-band count is null for CCD " << _ccdnr << "."
	<< Sas::message(Msg::App, Msg::Noisy);
    return - static_cast<double>(_counters[HARDBAND]);
  } // if (softband_count == 0ul)
  return  static_cast<double>(_counters[HARDBAND])
	/ static_cast<double>(softband_count     );
	// A single division is simple enough that we do not cache.
} // double AnomalousStateFinder::CCDdata::getH()

//------------------------------------------------------------------------------
/** Return the event rate, R = counts(RATEBAND) / GTIsum.
    This method recomputes R if increment(...)
    was called after the most recent call of getR().
    If there is no GTI for this CCD, a SAS error is issued. */
double AnomalousStateFinder::CCDdata::getR() {
  double GTIsum = getGTIsum();
  if (GTIsum == 0) {
    errstr << "No GTI for CCD " << _ccdnr << "." << Sas::error("noGTI");
    // Execution should not reach this point, but for elegance and robustness...
    GTIsum = -1.0;
  } // if (GTIsum == 0)
  return _counters[RATEBAND] / GTIsum;
	// A single division is simple enough that we do not cache.
} // double AnomalousStateFinder::CCDdata::getR()

//------------------------------------------------------------------------------
/** True if this CCD is in an anomalous state.
    This method recomputes the result if increment(...)
    was called after the most recent call of isAnomalous(). */
bool AnomalousStateFinder::CCDdata::isAnomalous() {
  // See if counters have been updated since last function call.
  bool cachedIsValid = true;
  for (unsigned i=0u; i < N_BANDS; i++)
    if (_counters[i] != _counters_last_accessed[i]) {
      cachedIsValid = false;
      _counters_last_accessed[i] = _counters[i];
    } // if (_counters[i] != _counters_last_accessed[i])

  // Recompute only if cached result is no longer valid.
  if (!cachedIsValid) {
    if (_evalHptrPos >= 0) *(_evalSyms.valuePtr[_evalHptrPos]) = getH();
    if (_evalRptrPos >= 0) *(_evalSyms.valuePtr[_evalRptrPos]) = getR();
    _isAnomalous = _evaluatorPtr->evalLogic();
  } // if (!cacheIsValid)
//if (_evalHptrPos >=0) errstr<<"H="<<getH()<<" ";
//if (_evalRptrPos >=0) errstr<<"R="<<getR()<<" ";
//errstr<<Sas::message(Msg::App,Msg::Always);

  return _isAnomalous;
} // bool AnomalousStateFinder::CCDdata::isAnomalous()

//------------------------------------------------------------------------------
/** Does the actual computation to sum all GTIs for this CCD. */
double AnomalousStateFinder::CCDdata::sumGTI() {
  // Construct the string ""STDGTI0n", nn == _ccdnr.
  ostringstream blockNameStream;
  blockNameStream << "STDGTI0" << _ccdnr;
  const string blockName = blockNameStream.str();

  // Check for the presence of a STDGTI0n Table.
  if ( !_dataSetPtr->hasBlock(blockName) ) {
    errstr << "The input FITS file does NOT contain a STDGTI extension for CCD "
	<< _ccdnr << "." << Sas::warning("noSTDGTIextension");
    return 0.0;
  } // if ( !_dataSetPtr->hasBlock(blockName) )

  // Read the content of the STDGTI0n Table from the DataSet.
  const Table * gtiTable = _dataSetPtr->table(blockName, Dal::Read);
  // STDGTI0n extensions are BINTABLEs with columns named START and STOP.
  const ColumnData * startData = gtiTable->column("START", Dal::Read)->data();
  const ColumnData *  stopData = gtiTable->column("STOP" , Dal::Read)->data();
  // See the Dal documentation for explanation of the following syntax.
  const real64 * start = startData->real64Data();
  const real64 * stop  =  stopData->real64Data();
  const unsigned long Nrows = gtiTable->column("START", Dal::Read)->rows();

  // Sum the GTIs.
  real64 GTIsum = (real64)0.0;
  for (unsigned long i=0ul; i < Nrows; i++)
    GTIsum += stop[i] - start[i];

  // See the Dal documentation for the need to delete the ColumnData instances.
  delete startData; startData = 0;
  delete  stopData;  stopData = 0;

  return (double)GTIsum;
} // double AnomalousStateFinder::CCDdata::sumGTI()

//------------------------------------------------------------------------------
unsigned long AnomalousStateFinder::getCount(const Ccd ccdnr, const CCDdata::EnergyBand band) {
  if (!_eventsRead) readEvents();
  return _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]->getCount(band);
} // unsigned long AnomalousStateFinder::getCount(const Ccd ccdnr, const CCDdata::EnergyBand band)

//------------------------------------------------------------------------------
double AnomalousStateFinder::getGTIsum(const Ccd ccdnr) {
  return _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]->getGTIsum();
} // double AnomalousStateFinder::getGTIsum(const Ccd ccdnr)

//------------------------------------------------------------------------------
double AnomalousStateFinder::getH(const Ccd ccdnr) {
  if (!_eventsRead) readEvents();
  return _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]->getH();
} // double AnomalousStateFinder::getH(const Ccd ccdnr)

//------------------------------------------------------------------------------
double AnomalousStateFinder::getR(const Ccd ccdnr) {
  if (!_eventsRead) readEvents();
  return _CCDdataPtrArray[ccdnr - EMOScam::CCD_FIRST]->getR();
} // double AnomalousStateFinder::getR(const Ccd ccdnr)

//------------------------------------------------------------------------------
/** Go through the EVENTS Table that _dataSetPtr points to,
    classify each event and increment the appropriate counter. */
void AnomalousStateFinder::readEvents() {
  _eventsRead = true;

  // Check for the presence of an EVENTS Table.
  if ( !_dataSetPtr->hasBlock("EVENTS") ) {
    Sas::warning("noEVENTSextension",
		"The input FITS file does NOT contain an EVENTS extension.");
    return;
  } // if ( !_dataSetPtr->hasBlock("EVENTS") )

  // Read the content of the EVENTS Table from the DataSet.
  const Table * eventTable = _dataSetPtr->table("EVENTS", Dal::Read);
  // EVENTS extensions are BINTABLEs; we are only interested in the columns
  // DETX, DETY and CCDNR, for `corner pixel' selection, and
  // PI, for calculating the hardness ratio H and count rate R.
  const ColumnData *  detXdata = eventTable->column("DETX" , Dal::Read)->data();
  const ColumnData *  detYdata = eventTable->column("DETY" , Dal::Read)->data();
  const ColumnData * ccdnrData = eventTable->column("CCDNR", Dal::Read)->data();
  const ColumnData *    piData = eventTable->column("PI"   , Dal::Read)->data();
  // See the Dal documentation for explanation of the following syntax.
  const unsigned long Nrows = eventTable->column("DETX", Dal::Read)->rows();
  const int16 *  detX =  detXdata->int16Data();
  const int16 *  detY =  detYdata->int16Data();
  const int16 *    pi =    piData->int16Data();
  const int8  * ccdnr = ccdnrData-> int8Data();

  // Classify the events and increment the appropriate counter.
  for (unsigned long i=0ul; i < Nrows; i++) {
    if ( !isCornerPixel(detX[i], detY[i])                  ) continue;
    unsigned ccdIndex = ccdnr[i] - EMOScam::CCD_FIRST;
    if ( _CCDdataPtrArray[ccdIndex]->isInBand(pi[i], CCDdata::RATEBAND) )
      _CCDdataPtrArray[ccdIndex]->increment(CCDdata::RATEBAND);
    if ( _CCDdataPtrArray[ccdIndex]->isInBand(pi[i], CCDdata::SOFTBAND) )
      _CCDdataPtrArray[ccdIndex]->increment(CCDdata::SOFTBAND);
    if ( _CCDdataPtrArray[ccdIndex]->isInBand(pi[i], CCDdata::HARDBAND) )
      _CCDdataPtrArray[ccdIndex]->increment(CCDdata::HARDBAND);
	// Note that the code segment above allows
	// for the overlapping of the hard and soft bands, and
	// for the non-inclusion of the hard and soft bands by the rate band,
	// at the expense of three extra logical operations per softband event.
  } // for (unsigned long i=0ul; i < Nrows; i++)

  // See the Dal documentation for the need to delete the ColumnData instances.
  delete  detXdata;  detXdata = 0;
  delete  detYdata;  detYdata = 0;
  delete ccdnrData; ccdnrData = 0;
  delete    piData;    piData = 0;
} // void AnomalousStateFinder::readEvents()

//------------------------------------------------------------------------------
bool AnomalousStateFinder::isCornerPixel(int16 detX, int16 detY) {
  if (_detXptrPos >= 0) *(_cornerEvalSyms.valuePtr[_detXptrPos]) = (int)detX;
  if (_detYptrPos >= 0) *(_cornerEvalSyms.valuePtr[_detYptrPos]) = (int)detY;
  return _cornerEvaluatorPtr->evalLogic();
} // bool AnomalousStateFinder::isCornerPixel(int16 detX, int16 detY)

// EOF

/*
$Log: AnomalousStateFinder.cc,v $
Revision 1.0  2009/04/16 09:38:32  cmhchen
Removed all hardcoded parameters; task now calls cal access methods instead.
class AnomalousStateFinder::CCDdata: New method isInBand(int, EnergyBand).

Revision 0.6  2009/04/15 10:01:52  cmhchen
Renamed package from goodpixelcount to emtaglenoise.
New methods to write LENOISnn as Attributes instead of printing to STDERR.
Removed all dependencies of class AnomalousStateFinder on class PixelCounter.

Revision 0.5  2009/01/22 13:33:41  cmhchen
Now tags verification states as anomalous as well.
New access methods getCount(Ccd, EnergyBand), getGITsum(Ccd), getH(Ccd)
 and getR(Ccd).
assert(...) are now Sas::error(...).

Revision 0.4  2008/10/27 19:45:38  cmhchen
Initial revision.

*/
