/****************************************************************
*
* AnomalousStateFInder.h	XMM Science Analysis System (SAS)
*
* Class containing routines to identify and tag anomalous states
* according to Kuntz and Snowden (2008) A&A 478, 575--596.
*
* (c) 2008 ESA - Written by Hubert Chen, CEA Saclay
*
* $Id: AnomalousStateFinder.h,v 1.0 2009/04/15 11:21:50 cmhchen Exp $
*
****************************************************************/

#ifndef ANOMALOUSSTATEFINDER_H
#define ANOMALOUSSTATEFINDER_H

// C++ global:
#include <sstream>
	// for ostringstream to convert integer indices to (sub)strings
	// to construct the FITS extension names STDGTI0n.
// #define NDEBUG
#include <cassert>
	// for debugging.  #define NDEBUG voids all assert(bool) calls.
// SAS global:
#include <param.h>
	// for command-line parameter access
	// param/src/param.h
#include <errstr.h>
	// specified in the `error' package C++ API, and includes <Error.h>
	// error/src/errstr.h
// SAS common:
#include <Xmm.h>
	// contains instrument parameter constants
	// caloalutils/src/Xmm.h
#include <CalServer.h>
	// to call the various methods of calServer.
	// cal/src/CalServer.h
#include <Dal.h>
	// for data access
	// dal/interface/Dal.h
#include <Evaluator.h>
	// for Evaluator class in selectlib.
	// selectlib/src/selcore/Evaluator.h
// Implementation specific:
// None so far.

/** Given the filename of a MOS merged event list (on the command line),
    decide if any CCD is in an anomalous state, and write the result
    as attributes to the input file header.
    This is the SAS task that gives access to the AnomalousStateFinder class. */
// void emtaganomalous();
	// defined in emtaganomalous.cc

/** Main class containing all methods for finding anomalous states.
    The static methods tagAnomalous(...) take care of everything. */
class AnomalousStateFinder {
	// defined in AnomalousStateFinder.cc
  public:
    /** Constructor that takes a filename string as input. */
    AnomalousStateFinder(const string filename);
    /** Constructor that takes a pointer to a DataSet instance as input. */
    AnomalousStateFinder(DataSet * dataSetPtr);

    ~AnomalousStateFinder();

    /** Main public interface with a filename string as input. */
    static void tagAnomalous(const string filename);
    /** Main public interface with a DataSet* as input. */
    static void tagAnomalous(DataSet * dataSetPtr);

    /** Wrapper to call writeAttribute(Ccd) for each CCD. */
    void AnomalousStateFinder::writeAttributes();
    /** Write anomalous-state attribute for ccdnr to _dataSet
	according to the return value of isAnomalous(ccdnr). */
    void AnomalousStateFinder::writeAttribute(Ccd ccdnr);

    /** True if the specified CCD is in an anomalous state. */
    bool isAnomalous(const Ccd ccdnr);

    /** Subroutine to provide abstraction for the `corner pixel' criteria. */
    const string cornerExpression();

    /** Return the selectlib expression that decides if the specified CCD is
	in an anomalous state.  The returned value contains no variable
	string other than "H" (hardness ratio) and "R" (count rate)
	as per Table 1 of Kuntz and Snowden (2008). */
    const string anomalousExpression(const Ccd ccdnr);

    /** Format string with which to construct the attribute names. */
    static const char * attributeNameFormat;
    /** Format string with which to construct the attribute comments. */
    static const char * attributeCommentFormat;
    /** Name of the attribute to write to DataSet. */
    static const string attributeName(Ccd ccdnr);
    /** Comment for the attribute to write to DataSet. */
    static const string attributeComment(Ccd ccdnr);

    /** Class to contain all data pertaining to a CCD.
	Callers instantiate the class with the CCDNR, and
	increment event counters through the method increment(...).
	When event accumulation is complete, various get* methods
	return saved quantities, and isAnomalous() returns the decision. */
    class CCDdata {
      public:
	/** Constructor.
	    DataSet must be of instrument Xmm::EMOS1 or XMM::EMOS2.
	    ccdnr is the CCD number of this CCDdata.
	    Ccd is defined in caloalutils/src/Xmm.h .
	    DataSet is defined in dal/interface/Dal.h . */
	CCDdata(const DataSet * _dataSetPtr, const Ccd ccdnr);
	~CCDdata();

	/** Return the selectlib expression that decides if this CCD is
	in an anomalous state.  The returned value contains no variable
	string other than "H" (hardness ratio) and "R" (count rate)
	as per Table 1 of Kuntz and Snowden (2008). */
	const string anomalousExpression();

	/** Identifiers for accessing private counters
	    in an implementation-independent manner. */
	enum EnergyBand { HARDBAND, SOFTBAND, RATEBAND, N_BANDS };

	/** Decide if eVtestEnergy, in eV, is within the EnergyBand band. */
	bool isInBand(int eVtestEnergy, const EnergyBand band) const;

	// set methods
	/** Increment the event counter for the specified band. */
	void increment(const EnergyBand band, const unsigned long increment = 1ul);

	// access methods
	/** Return the event count (to date) for the specified band. */
	unsigned long getCount(const EnergyBand band) const;
	/** Return the sum of the GTIs for this (Instrument,Ccd). */
	double getGTIsum();
	/** Return the hardness ratio, H = counts(HARDBAND) / counts(SOFTBAND).
	This method recomputes H if increment(...)
	was called after the most recent call of getH().
	If the soft-band count is null,
	a warning is issued, and -1 * hard-band count is returned. */
	double getH();
	/** Return the event rate, R = counts(RATEBAND) / GTIsum.
	This method recomputes R if increment(...)
	was called after the most recent call of getR().
	If there is no GTI for this CCD, a SAS error is issued. */
	double getR();
	/** True if this CCD is in an anomalous state.
	This method recomputes the result if increment(...)
	was called after the most recent call of isAnomalous(). */
	bool isAnomalous();

      protected:
	/** Does the actual computation to sum all GTIs for this CCD. */
	double sumGTI();

      private:
	// input
	const DataSet * _dataSetPtr;
	Instrument	_instrument;
	const Ccd	_ccdnr;		// for retrieving CCD ID specific data.
	// internal
	unsigned long	_counters[N_BANDS],
			_counters_last_accessed[N_BANDS];
				// to check the validity of cached output.
	int		_minEnergy[N_BANDS],
			_maxEnergy[N_BANDS];
	// selectlib infrastructure for evaluating isAnomalous().
	// The following variables are from selectlib/src/selcore/Evaluator.h .
	Evaluator *	_evaluatorPtr;
	NumericSymbols	_evalSyms;
	int		_evalHptrPos, _evalRptrPos;
	// output
	double		_GTIsum;	// cache to save recomputing
	bool		_isAnomalous;	// cache to save recomputing
    }; // class AnomalousStateFinder::CCDdata

  // access methods
  unsigned long getCount(const Ccd ccdnr, const CCDdata::EnergyBand band);
  double getGTIsum(const Ccd ccdnr);
  double getH(const Ccd ccdnr);
  double getR(const Ccd ccdnr);

  protected:
    /** Go through the EVENTS Table that _dataSetPtr points to,
	classify each event and increment the appropriate counter. */
    void readEvents();
    /** Abstraction for the exact implementation for the criteria. */
    bool isCornerPixel(int16 detX, int16 detY);

  private:
    /** Handle tasks common to all constructors. */
    void init();

    // input
    const string _filename; // kept as a flag to decide whether to close file.
    DataSet * _dataSetPtr;
			// the DataSet with which a class instance is associated
    // internal
    CCDdata **		_CCDdataPtrArray;
    bool _eventsRead;	// to remember if readEvents() has been called.
    // selectlib infrastructure for evaluating isCornerPixel().
    // The following variables are from selectlib/src/selcore/Evaluator.h .
    Evaluator *		_cornerEvaluatorPtr;
    NumericSymbols	_cornerEvalSyms;
    int			_detXptrPos, _detYptrPos;
}; // class AnomalousStateFinder

#endif
// ifndef ANOMALOUSSTATEFINDER_H

/*
$Log: AnomalousStateFinder.h,v $
Revision 1.0  2009/04/15 11:21:50  cmhchen
Removed all hardcoded parameters; task now calls cal access methods instead.
class AnomalousStateFinder::CCDdata: New method isInBand(int, EnergyBand).

Revision 0.6  2009/04/15 10:02:11  cmhchen
Renamed package from goodpixelcount to emtaglenoise.
New methods to write LENOISnn as Attributes instead of printing to STDERR.
Removed all dependencies of class AnomalousStateFinder on class PixelCounter.

Revision 0.5  2008/10/28 16:04:30  cmhchen
New access methods getCount(Ccd, EnergyBand), getGITsum(Ccd), getH(Ccd)
 and getR(Ccd).
Moved '#include "emtaganomalous.h"' macro to emtaglenoise.cc .

Revision 0.4  2008/10/27 19:48:14  cmhchen
Initial revision.

*/
