/********************************************************
*
* PixelCounter.h	XMM Science Analysis System (SAS)
*
* Class containing routines to calculate and write no. of
* good and selected MOS pixels to DataSets.
*
* (c) 2008 ESA - Written by Hubert Chen, CEA Saclay
*
* $Id: PixelCounter.h,v 0.5 2008/10/28 15:06:39 cmhchen Exp $
*
********************************************************/

#ifndef GOODPIXELCOUNT_H
#define GOODPIXELCOUNT_H

// C++ global:
#include <sstream>
	// for ostringstream to convert integer indices to (sub)strings
// #define NDEBUG
#include <cassert>
	// for debugging.  #define NDEBUG voids all assert(bool) calls.
// SAS global:
#include <param.h>
	// for command-line parameter access
	// param/src/param.h
#include <errstr.h>
	// specified in the `error' package C++ API, and includes <Error.h>
	// error/src/errstr.h
// SAS common:
#include <Xmm.h>
	// contains instrument parameter constants
	// caloalutils/src/Xmm.h
#include <CalServer.h>
	// to call the various methods of calServer.
	// cal/src/CalServer.h
#include <Dal.h>
	// for data access
	// dal/interface/Dal.h
#include <Evaluator.h>
	// for Evaluator class in selectlib.
	// selectlib/src/selcore/Evaluator.h
// #include <DataSubspace.h>
	// dsslib/src/dsscore/DataSubspace.h
// Implementation specific:
#include <ShapeScanner.h>
	// for RegionScanner and PolygonScanner; includes scanningLib.h .
	// rgslib/lib/ShapeScanner.h
#include <PositionUtils.h>
	// high-abstraction conversion between RAWX/Y and DETX/Y
	// arfgen/src/PositionUtils.h

/** Given the filename of an event list (on the command line),
 compute the number of good pixels on each CCD, based on the content
 of the BADPIXnn extensions.
(NOT IMPLEMENTED: and if a DSS extension is present,
 include only pixels within the selection.)
Optional: Provide a selectlib-style expression for extra filtering.
Specify corner=yes to filter according to the `corner pixel' criteria
 in Table 3 of Kuntz and Snowden (2008), A&A 478, 575--596.
If both expression and corner are not provided (or is null),
 the pixel counts are written to the file header as values of keyword
 attributes NGOODPIn (n=1..7) for each CCD and NGOODPIX for their sum.
Otherwise, the numbers are only printed to STDOUT to preserve the integrity
 of all attributes in the input file.
This is an XMMSAS task that reads parameters via the common mechanism. */
void goodpixelcount();
	// defined in goodpixelcount.cc

/** Main class containing all methods for counting pixels. */
class PixelCounter {
	// defined in PixelCounter.cc
  public:
    /** Constructor that takes a filename string as input. */
    PixelCounter(const string filename);
    /** Constructor that takes a pointer to a DataSet instance as input. */
    PixelCounter(DataSet * dataSetPtr);

    ~PixelCounter();

    /** Static function to calculate and add NGOODPIn (n=1..7)
	for each MOS CCD and NGOODPIX for their sum as keyword
	attributes to the FITS file specified by filename. */
    static void writeGoodPixelCount(	const string filename);
    /** Static function to calculate and print to STDOUT no. of good
	and selected pixels, based on the DSS in the FITS filename
	and on the selectlib expression. */
    static void printGoodPixelCount(	const string filename,
					const string expression,
					const bool cornerFlag = false);

    /** Return in a vector the number of good and selected pixels
	in all MOS CCDs. */
    vector<unsigned long> goodPixelCounts();
    /** Return the number of good and selected pixels in the MOS CCD
	whose ID is ccdnr (type Ccd defined in caloalutils/src/Xmm.h). */
    unsigned long goodPixelCount(const Ccd ccdnr);

    /** Remember the new selectlib expression expr in state variable(s).
	expr will be AND-ed with any existing selection criteria. */
    void addExpression(const string expr);
    /** Subroutine to provide abstraction for the `corner pixel' criteria. */
    static const string cornerExpression(Instrument instrument);
	// Instrument type defined in caloalutils/src/Xmm.h

    /** Return in a vector the number of good PIXCOORD1 pixels
	(RAWX/Y) in all MOS CCDs.
	This is based purely on the BADPIXnn extensions; any DSS
	or input selectlib expression is not taken into account. */
    vector<unsigned long> goodRawPixelCounts();
    /* Ditto, for the specified ccdnr. */
    unsigned long goodRawPixelCount(const Ccd ccdnr);
    /** Return pixel maps, one for each MOS CCD, whose values are
	false at bad pixels and true otherwise. */
    bool *** getGoodPixelMaps();
    /** Ditto, for the specified ccdnr. */
    bool ** getGoodPixelMap(const Ccd ccdnr);

    /** Corner pixel specification per
	Table 3 in K. D. Kuntz and S. L. Snowden (2008), A&A 478, 575--596. */
    static const string EMOS1cornerExpression;
    static const string EMOS2cornerExpression;

  protected:
    /** Calculate sum( sum(_goodPixelMaps[ccdnr] )) */
    unsigned long sumGoodPixelMap(const Ccd ccdnr);
    /** Read the extension BADPIX0n (n=ccdnr) from DataSet,
	compute and return the map of good pixels on the given CCD. */
    bool ** computeGoodPixelMap(const Ccd ccdnr);

//    void filterRegion(GeoRegion region);
//  DataSubspace DSS_from_DataSet(DataSet dataSet);
//  string ANDexpressions(const string expr1, const string expr2);
    /** A not-so-trivial class to produce a list<Vertex>
	for the corners of the named CCD in CAMCOORD2 (DETX/Y),
	for feeding into PolygonScanner::(list<scanningLib::Vertex>). */
    list<scanningLib::Vertex> CCDcorners(Ccd ccdnr);
//  GeoRegion
//    static bool hasDSS(const DataSet & dataSet);
//    static string DSS_to_expr(...);
//    static void det_to_raw(...);

    /** wrapper around filterCCD{fine,coarse}(Ccd ccdnr) */
    unsigned long filterCCD(Ccd ccdnr, bool fine=0);
    /** Compute the number of good and selected RAWX/Y pixels on CCD ccdnr. */
    unsigned long filterCCDcoarse(Ccd ccdnr);
    /** Compute the number of good and selected DETX/Y pixels on CCD ccdnr.
	Caution: This routine runs *forever*. */
    unsigned long filterCCDfine  (Ccd ccdnr);
    unsigned long filterRegion(scanningLib::PolygonScanner polygonScanner);

  private:
    void init();	// to handle tasks common to all constructors
    vector<scanningLib::Vertex> CCDcorners();

    //  input
    DataSet * _dataSetPtr;	// kept just to close the file
    string _expr;

    // output
    /** Maps registering which MOS pixels are good (ie, not bad).
	Dimensions == CCDNR x n_RAWX x n_RAWY */
    bool *** _goodPixelMaps;
    /** sum( sum(_goodPixelMaps )), dimension == CCDNR, */
    long * _goodPixelCounts;
	// remembered to save calculation in subsequent calls.
    long * _selectedPixelCounts;
	// remembered to save calculation in subsequent calls.
    bool * _selectionValidity;

    // internal
    /** Whether ***_goodPixelMaps and *_goodPixelCounts are still valid.
	Set after calculation; unset after addExpression(). */
    bool * _status;
    /** Local placeholders for calServer.instrument()->{min,max}{X,Y}() */
    unsigned _rawXmin, _rawXmax, _rawYmin, _rawYmax, _nCcds;

}; // class PixelCounter

#endif
// ifndef GOODPIXELCOUNT_H

/*
$Log: PixelCounter.h,v $
Revision 0.5  2008/10/28 15:06:39  cmhchen
Moved '#include "goodpixelcount.h"' macro to goodpixelcount.cc .

Revision 0.4  2008/10/25 23:00:45  cmhchen
First revision with a working PixelCounter and task goodpixelcount.
New class AnomalousStateFinder and task emtaganomalous.

Revision 0.3  2008/10/06 18:26:20  cmhchen
This revision accepts a user-input selectlib-style expression for filtering,
in addition to discounting bad pixels based on the BADPIX0n extensions.
It counts in integral DETX/Y coordinates, covering all pixels on any CCD.
Works in principle, but looping through all 7 x 12000^2 DETX/Y takes forever.

Revision 0.2  2008/09/09 15:56:36  cmhchen
Restructured code from previous version to formal class PixelCounter.
GoodPixelCount.h now becomes PixelCounter.h .
There should be no change in functionality whatsoever.

Revision 0.1  2008/09/05 17:56:44  cmhchen
First sandbox version that does bad pixel filtering based on BADPIXnn.
Pixel counts are written to file header in PIXCOORD1 (RAWX/Y).

*/
